package de.uks.plm.ftglb.incremental;
/**
 * <p>
 * A TaskQueue[Queue, Z] (Z is the type of the result, Queue is the concrete 
 * type of TaskQueue).
 * The way as it is today (with Queue as one of the required generic types) 
 * is because performance issue, i.e., compiler can replace TaskQueue with Queue
 * at the compile time, thus saving a virtual function loop up whenever it is 
 * called within GLB. No semantic benefits.
 * Represents the user-defined data-structure maintained at each place by the
 * GLB implementation. It is recommended to have an associated {@link TaskBag}
 * to facilitate the split/merge methods. 
 */
public interface TaskQueue[Queue, R] // {Queue<:TaskQueue[Queue, R]}
{
    //public abstract def process(n:Long, context:ContextI):Boolean;

    public abstract def process(n:Long) {Queue<:TaskQueue[Queue, R]} : Boolean;

    /**
     * Split the current TaskBag
     * @return null if TaskBag is too small to return
     *         TaskBag split
     */
    public def split() : TaskBag;

    /**
     * Merge TaskBag at the top of the queue.
     */
    public def merge(val tb:TaskBag) : void;
    // public def merge(val that:Queue) : void;

    /**
     * Returns the number of task items that have been prcoessed.
     * This method is used to keep track of task item statistics (i.e. a hint),
     * whether it is implemented faithfully is not important.
     */
    public  def count() : Long;

    /**
     * Returns the computation result.
     * @return computation result
     */
    public def getResult() : GLBResult[R];

    /**
     * Auxiliary function that user can overwrite to print the statistics one
     * cares about, e.g., computation time. This function will be called only
     * after all the calculation and results reduction are done.
     */
    public  def printLog() : void;

    /**
     * Get the size of the queue.
     */
    public def size() : Long;

    /**
     * Make a (deep!) clone of this queue.
     */
    public def clone() : Queue;

    /**
     * @param n elements to get from the top of the queue
     * @result a queue with n elements
     */
    public def getQueueFromTop(val n:Long) : Queue;

    /**
     * @param n elements to get from the bottom of the queue
     * @result a queue with n elements
     */
    public def getQueueFromBottom(val n:Long) : Queue;

    /**
     * Get the top element, as well as the result as a new queue. The top 
     * element is only copied, not popped.
     */
    public def getTopElement() : Queue;

    /**
     * removes n elements from the top of this queue
     * 
     * @param n how much elements
     */
    public def removeFromTop(val n:Long) : void;

    /**
     * removes n elements from the bottom of this queue
     * 
     * @param n how much elements
     */
    public def removeFromBottom(val n:Long) : void;

    /**
     * Merge Queue at the top of current Queue , thus changing the state of 
     * the current Queue. The results are NOT changed.
     *
     * @param q incoming Queue
     */
    public def mergeAtTop(val that:Queue) : void;

    /**
     * Merge Queue at the bottom of current Queue , thus changing the state of
     * the current Queue. The results are NOT changed.
     *
     * @param q incoming Queue
     */
    public def mergeAtBottom(val that:Queue) : void;

    /**
     * we need a possibility to overwrite the result-Rail.
     */
    public def setResult(val that:Queue) : void;

    /**
     * Merge the result of that into the result of this queue.
     * @param that the other queue.
     */
    public def mergeResult(val that:Queue) : void;
}