package de.uks.plm.ftglb.incremental.bcgOneQueue;

import x10.compiler.Inline;

import x10.util.Pair;
import x10.util.Random;
import x10.util.Team;

import de.uks.plm.ftglb.incremental.GLBResult;
import de.uks.plm.ftglb.incremental.TaskBag;
import de.uks.plm.ftglb.incremental.TaskQueue;

import de.uks.plm.utils.AssertPrinter;

import de.uks.plm.utils.RingBuffer;

public class Queue implements TaskQueue[Queue, Double]
{
    private val graph:Graph;

    private var data:Rail[Int];
    private var size:Long;

    private val regularQueue:FixedRailQueue[Int];
    private val verticesToWorkOn:Rail[Int];
    private var betweennessMap:Rail[Double];
    private val realBetweennessMap:Rail[Double];
    private val distanceMap:Rail[Long];
    private val sigmaMap:Rail[Long];
    private val deltaMap:Rail[Double];
    private var count:Int;
    private val n:Int;
    private val m:Int;
    private var state:Int;
    private var s:Int;
    val predecessorMap:Rail[Int];
    val predecessorCount:Rail[Int];

    public def this(val rmat:Rmat, val permute:Int) : Queue
    {
        this.graph = rmat.generate();
        this.graph.compress();
        this.n = graph.numVertices();
        this.m = graph.numEdges();
        this.regularQueue = new FixedRailQueue[Int](n);

        val initVertices:(Long) => Int = (val n:Long) =>
                                         {
                                             return (n as Int);
                                         };

        this.verticesToWorkOn = new Rail[Int](n, initVertices);

        if (permute > 0)
        {
            this.permuteVertices();
        }
        this.betweennessMap = new Rail[Double](n, 0);
        this.realBetweennessMap = new Rail[Double](n, 0);
        this.distanceMap = new Rail[Long](n, Long.MAX_VALUE);
        this.sigmaMap = new Rail[Long](n, 0);
        this.deltaMap = new Rail[Double](n, 0);
        this.predecessorMap = new Rail[Int](graph.numEdges());
        this.predecessorCount = new Rail[Int](n);
        this.state = 0n;
        this.s = 0n;
        val max:Long = Place.numPlaces();
        val h:Long = here.id();
        val lower:Long = (n as Long * h / max) as Int;
        val upper:Long = (n as Long * (h + 1) / max) as Int;
        this.size = upper - lower;
        
        val init:(Long) => Int = (val index:Long) =>
                                 {
                                     return ((index + lower) as Int);
                                 };

        this.data = new Rail[Int](size, init);
    }

    public def this(val n:Int) : Queue
    {
        this(null as Queue, n);
    }

    public def this(val that:Queue, val n:Int) : Queue
    {
        this.n = n;
        this.m = 0n;
        this.regularQueue = (null as FixedRailQueue[Int]);
        this.verticesToWorkOn = null;
        this.realBetweennessMap = null;
        this.distanceMap = null;
        this.sigmaMap = null;
        this.deltaMap = null;
        this.graph = null;
        this.predecessorMap = null;
        this.predecessorCount = null;
        this.state = 0n;
        this.s = 0n;
        if (null != that)
        {
            this.size = that.size;
            this.data = new Rail[Int](that.data);
            this.betweennessMap = new Rail[Double](that.betweennessMap);
        }
        else
        {
            this.size = 0;
            this.data = null;
            this.betweennessMap = new Rail[Double](n);
        }
    }

    @Inline protected final def bfsShortestPath1(val s:Int) : void
    {
         // Put the values for source vertex
        this.distanceMap(s) = 0;
        this.sigmaMap(s) = 1;
        this.regularQueue.push(s);
    }

    @Inline protected final def bfsShortestPath2() : void
    {
        this.count++;
        // Pop the node with the least distance
        val v = this.regularQueue.pop();

        // Get the start and the end points for the edge list for "v"
        val edgeStart:Int = graph.begin(v);
        val edgeEnd:Int = graph.end(v);

        // Iterate over all its neighbors
        for (var wIndex:Int = edgeStart; wIndex < edgeEnd; ++wIndex)
        {
            // Get the target of the current edge.
            val w:Int = graph.getAdjacentVertexFromIndex(wIndex);
            val distanceThroughV:Long = this.distanceMap(v) + 1;

            // In BFS, the minimum distance will only be found once --- the 
            // first time that a node is discovered. So, add it to the queue.
            if (distanceMap(w) == Long.MAX_VALUE)
            {
                this.regularQueue.push(w);
                this.distanceMap(w) = distanceThroughV;
            }

            // If the distance through "v" for "w" from "s" was the same as its 
            // current distance, we found another shortest path. So, add 
            // "v" to predecessorMap of "w" and update other maps.
            if(distanceThroughV == this.distanceMap(w))
            {
                this.sigmaMap(w) = this.sigmaMap(w) + this.sigmaMap(v);// XTENLANG-2027
                val idx:Long = this.graph.rev(w) + (this.predecessorCount(w)++);
                this.predecessorMap(idx) = v;
            }
        }
    }

    @Inline protected final def bfsShortestPath3() : void
    {
        this.regularQueue.rewind();
    }

    @Inline protected final def bfsShortestPath4(val s:Int) : void
    {
        val w = this.regularQueue.top();
        val rev = this.graph.rev(w);
        while (this.predecessorCount(w) > 0)
        {
            val v = this.predecessorMap(rev + (--this.predecessorCount(w)));

            this.deltaMap(v) +=  (  (this.sigmaMap(v) as Double) 
                                  / (this.sigmaMap(w) as Double))
                               * (1.0 + this.deltaMap(w));
        }

        // Accumulate updates locally 
        if(w != s)
        {
            this.realBetweennessMap(w) += this.deltaMap(w); 
        }
        this.distanceMap(w) = Long.MAX_VALUE;
        this.sigmaMap(w) = 0;
        this.deltaMap(w) = 0.0;
    }

    @Inline public def clone() : Queue
    {
        return new Queue(this, this.n);
    }

    @Inline public def count() : Long
    {
        return (this.count);
    }

    // @override
    public def getResult() : BCGResult
    {
        val result = new BCGResult();
        return (result);
    }

    public def getQueueFromBottom(val n:Long) : Queue
    {
        val get:Long = (n > this.size ? this.size : n);
        val queue = new Queue(this.n);
        val data:Rail[Int] = new Rail[Int](get);
        Rail.copy(this.data, 0, data, 0, get);
        queue.data = data;
        queue.size = get;
        return (queue);
    }

    public def getTopElement() : Queue
    {
        val queue = new Queue(this.n);
        val data:Rail[Int] = new Rail[Int](1, this.data(this.size - 1));
        queue.data = data;
        queue.betweennessMap = this.betweennessMap;
        queue.size = 1;
        return (queue);
    }

    public def grow() : void
    {
        val newSize:Long = Math.max(this.data.size * 2, 1);
        val newData:Rail[Int] = new Rail[Int](newSize);
        Rail.copy(this.data, 0, newData, 0, this.size);
        this.data = newData;
    }

    public def mergeAtBottom(val that:Queue) : void
    {
        if ((null == that) || (0 == that.size()))
        {
            return;
        }
        val newSize:Long = this.size + that.size;
        while (newSize > this.data.size)
        {
            this.grow();
        }
        Rail.copy(this.data, 0, this.data, that.size, this.size);
        Rail.copy(that.data, 0, this.data, 0, that.size);
        this.size = newSize;
    }

    public def mergeAtBottom(val bag:TaskBag) : void
    {
        this.mergeAtBottom(bag as Bag);
    }

    public def mergeAtBottom(val bag:Bag) : void
    {
        if ((null == bag) || (0 == bag.size()))
        {
            return;
        }
        val bagSize:Long = bag.size();
        val newSize = this.size + bagSize;
        while (newSize > this.data.size)
        {
            this.grow();
        }
        Rail.copy(this.data, 0, this.data, bagSize, this.size);
        Rail.copy(bag.data, 0, this.data, 0, this.size);
        this.size = newSize;
    }

    private def protResults() : void
    {
        this.betweennessMap = new Rail[Double](this.realBetweennessMap);
    }

    public def mergeResult(val that:TaskQueue[Queue, Double]) : void
    {
        assert (false)
                :  here 
                 + ": in mergeResult(val that:TaskQueue[Queue, Long]): "
                 + "Parameter is not of type Queue!";
    }
    public def mergeResult(val that:Queue) : void
    {
        assert (this.betweennessMap.size == that.betweennessMap.size)
               : here + ": cannot set result, rails do not match in length!";

        for (index:Long in 0..(this.betweennessMap.size - 1))
        {
            this.betweennessMap(index) += that.betweennessMap(index);
            if (null != this.realBetweennessMap)
            {
                this.realBetweennessMap(index) += that.betweennessMap(index);
            }
        }
    }

    /**
     * A function to shuffle the vertices randomly to give better work dist.
     */
    private def permuteVertices()
    {
        val prng = new Random(1);

        for(var i:Int = 0n; i < this.n; i++)
        {
            val indexToPick = prng.nextInt(this.n - i);
            val v = verticesToWorkOn(i);
            this.verticesToWorkOn(i) = this.verticesToWorkOn(i+indexToPick);
            this.verticesToWorkOn(i+indexToPick) = v;
        }
    }

    public def printLog() : void
    {
        Console.OUT.println(  "[" + here + "]"
                            // + " Time = " + accTime
                            + " Count = " + count);
    }

    @Inline public def process(val n:Long) : Boolean
    {
        var i:Long = 0;
        switch (state)
        {
            case 0n:
                // val top = size - 1;
                // val l = lower(top);
                // val u = upper(top) - 1n;
                // 
                // if(u == l) 
                // {
                //     --size;
                // }
                // else
                // {
                //     upper(top) = u;
                // }
                val u:Int = this.data(this.size - 1);
                // refTime = System.nanoTime();
                this.s = this.verticesToWorkOn(u);
                this.state = 1n;
            case 1n:
                bfsShortestPath1(s);
                this.state = 2n;
            case 2n:
                while(!regularQueue.isEmpty())
                {
                    if (i++ > n)
                    {
                        return (true);
                    }
                    this.bfsShortestPath2();
                }
                this.state = 3n;
            case 3n:
                this.bfsShortestPath3();
                this.state = 4n;
            case 4n:
                while(!regularQueue.isEmpty())
                {
                    if (i++ > n)
                    {
                        return (true);
                    }
                    this.bfsShortestPath4(s);
                }
                // accTime += (System.nanoTime()-refTime)/1e9;
                --this.size;
                this.protResults();
                this.state = 0n;
            }
        return (0 < this.size());
    }

    public def removeFromTop(val n:Long) : void
    {
        this.size -= n;
        if (this.size < 0)
        {
            this.size = 0;
        }
    }

    public def removeFromBottom(val n:Long) : void
    {
        val remove = (n > this.size ? this.size : n);
        this.size -= remove;
        Rail.copy(this.data, remove, this.data, 0, this.size);
    }

    public def setResult(val result:BCGResult) : void
    {
        assert (this.realBetweennessMap.size == result.getResult().size)
               : here + ": cannot set result, rails do not match in length!";

        this.betweennessMap = result.getResult();
    }

    public def setResult(val that:Queue) : void
    {
        assert (this.betweennessMap.size == that.betweennessMap.size)
               : here + ": cannot set result, rails do not match in length!";

        this.betweennessMap = that.betweennessMap;
    }

    public def size() : Long
    {
        return (this.size);
    }

    public def split() : Bag
    {
        if (2 > this.size())
        {
            return (null);
        }
        val otherHalf:Long = this.size / 2;
        val myHalf:Long = this.size - otherHalf;
        val lootData:Rail[Int] = new Rail[Int](otherHalf);
        Rail.copy(this.data, 0, lootData, 0, otherHalf);
        Rail.copy(this.data, otherHalf, this.data, 0, myHalf);
        this.size = myHalf;
        val loot:Bag = new Bag(lootData);
        return (loot);
    }
    /**
     * substring helper function
     */
    public static def sub(val str:String, val start:Int, val end:Int)
    {
        return (str.substring(start, Math.min(end, str.length())));
    }

    public class BCGResult extends GLBResult[Double]
    {
        // public val r:Rail[Double];
        // 
        // public def this(val r:Rail[Double]) : BCGResult
        // {
        //     this.r = new Rail[Double](r);
        // }

        public  def getResult() : Rail[Double]
        {
            return (betweennessMap);
        }

        public  def getReduceOperator() : Int
        {
            return (Team.ADD);
        }
    
        public def display(val r:Rail[Double]) : void
        {
            for (var i:Int = 0n; i < n; ++i)
            {
                if (0.0 != r(i))
                {
                    Console.OUT.println(  "(" + i + ") -> "
                                        + sub("" + r(i), 0n, 6n));
                }
            }
        }
    }
}