package de.uks.plm.ftglb.incremental.managers;

import x10.compiler.Inline;
import x10.compiler.NonEscaping;

public class BackForthManager
{
    /**
     * id of the place, we back up
     */
    private var forth:Long;

    /**
     * id of the place, we are backed up by
     */
    private var back:Long;

    public def this()
    {
        this.forth = -1;
        this.back = -1;
        this.updateBack();
        this.updateForth();
    }

    @Inline final public def resetForth() : void
    {
        this.forth = -1;
    }

    /**
     * Update back and forth.
     * Look for the next back- and forth-places, that are still alive.
     * Role: any
     * Sink: yes
     */
    @NonEscaping @Inline final public def updateBack() : void
    {
        var p:Place = Place.places().next(here);
        if (-1 != this.back)
        {
            p = Place(this.back);
        }
        while (p.isDead())
        {
            p = Place.places().next(here);
        }
        this.back = p.id();
    }

    @NonEscaping @Inline final public def updateForth() : void
    {
       var p:Place = Place.places().prev(here);
       if (-1 != this.forth)
       {
           p = Place(this.forth);
       }
       while (p.isDead())
       {
           p = Place.places().prev(here);
       }
       this.forth = p.id();
    }

    /**
     * Getter for back.
     */
    @Inline public def getBack() : Long
    {
        return (this.back);
    }

    /**
     * Getter for forth.
     */
    @Inline public def getForth() : Long
    {
        return (this.forth);
    }
}