package de.uks.plm.ftglb.incremental.managers;

import x10.compiler.Inline;
import x10.compiler.Uncounted;

import x10.util.ArrayList;
import x10.util.HashMap;
import x10.util.List;
import x10.util.Map;
import x10.util.Pair;
import x10.util.Set;

import de.uks.plm.ftglb.incremental.TaskQueue;
import de.uks.plm.ftglb.incremental.Worker;

import de.uks.plm.utils.AssertPrinter;

/**
 * Manage the links when i am acting as backup-place for someone else.
 */
public class LinkManager
{

    /**
     * Key = place-id
     * value = list of tans
     */
    public val links:Map[Long, ArrayList[Long]];

    public def this()
    {
        this.links = new HashMap[Long, ArrayList[Long]]();
    }

    /**
     * Add a link
     * Role: back(thief)
     * sink: yes
     */
    @Inline public def add(val tan:Long, val placeId:Long) : void
    {
        val entry:Pair[Long, Long] = new Pair[Long, Long](placeId, tan);
        this.add(entry);
    }

    /**
     * Add a link
     * Role: back(thief)
     * sink: yes
     */
    @Inline public def add(val entry:Pair[Long, Long]) : void
    {
        val placeId:Long = entry.first;
        val tan:Long = entry.second;

        if (this.links.containsKey(placeId))
        {
            this.links.get(placeId).add(tan);
        }
        else
        {
            val list:ArrayList[Long] = new ArrayList[Long]();
            list.add(tan);

            this.links.put(placeId, list);
        }
        AssertPrinter.println("[LinkMananger " + here + "]: " + this.links.get(placeId));
    }

    @Inline public def isEmpty() : Boolean
    {
        val places:Set[Long] = this.links.keySet();
        for (val place:Long in places)
        {
            if (!this.links.get(place).isEmpty())
            {
                return (false);
            }
        }
        return (true);
    }

    /**
     * Remove a link
     * Role: back(thief)
     * sink: yes
     */
    @Inline public def remove(val tan:Long, val placeId:Long) : void
    {
        val entry:Pair[Long, Long] = new Pair[Long, Long](placeId, tan);
        this.remove(entry);
    }

    /**
     * Remove a link
     * Role: back(thief)
     * sink: yes
     */
    @Inline public def remove(val entry:Pair[Long, Long]) : void
    {
        val list:List[Long] = this.links.get(entry.first);
        if (null != list)
        {
            list.remove(entry.second);
        }
    }
}