package de.uks.plm.ftglb.incremental.managers;

import x10.compiler.Inline;
import x10.compiler.Uncounted;

import de.uks.plm.ftglb.incremental.TaskQueue;
import de.uks.plm.ftglb.incremental.Worker;

/**
 * This class manages (cyclic) monitoring of places. To get the next place,
 * a reference to the corresponding worker is needed. Parameters Queue and R
 * are the same as for the worker.
 */
public class MonitorManager[Queue, R] {Queue <: TaskQueue[Queue, R]}
{
    /**
     * Reference to the corresponding
     * @link{de.uts.plm.ftglb.incremental.Worker worker}.
     */
    public val worker:Worker[Queue, R];

    /**
     * Constructor.
     *
     * @param worker the corresponding
     * @link{de.uts.plm.ftglb.incremental.Worker worker}.
     */
    public def this(val worker:Worker[Queue, R])
    {
        this.worker = worker;
    }

    // /**
    //  * Method to start the monitoring-process.
    //  */
    // @Inline public def monitor():void
    // {
    //     if (!this.worker.state.active)
    //     {
    //         this.worker.processRecorded();
    //         val backId:Long = this.worker.backForth.getBack();
    //         val back:Place = Place(backId);
    //         if (!back.isDead())
    //         {
    //             val st:PlaceLocalHandle[Worker[Queue, R]] = this.worker.st;
    //             at (back) async
    //             {
    //                 st().monitor();
    //             }
    //         }
    //         else
    //         {
    //             // TODO: restore
    //         }
    //     }
    // }

    /**
     * This method checks the next place according to the cyclic scheme is 
     * alive. If the place is not alive, restore is started and this process
     * waits, until restore is finished and rechecks the (new) backup-place.
     * When the check is finally successful,
     * @link{#monitor(Long) mointor(startedBy)} is called on the next place.
     */
    @Inline private def next(val startedBy:Long)
    {
        if (startedBy != this.worker.me)
        {
            val backId:Long = this.worker.backForth.getBack();
            val back:Place = Place(backId);
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.worker.st;
            if (!back.isDead())
            {
                at (back) async
                {
                    if (!st().state.active)
                    {
                        st().processRecorded();
                        st().monitor.monitor(startedBy);
                    }
                }
            }
            else // if (back.isDead())
            {
                val backBack:Place = Place.places().next(back);
                val backBackId:Long = backBack.id();

                /*
                 * If i am the backup-place of my backup-place and my back-place
                 * dies, then i am the last place alive. In this case no
                 * restore is necessary.
                 */
                if (backBackId != this.worker.me)
                {
                    this.worker.isDead(backBackId);
                }
            }
        }
    }

    /**
     * This method calls @link{#next(Long) next(started by) iff its place-id
     * is not equal to staredBy (since if it is equal, one round-trip is over).
     *
     * @param startedBy the id of the place, which started the monitor-process.
     */
    public def monitor(val startedBy:Long):void
    {
        if (startedBy != this.worker.me)
        {
            this.next(startedBy);
        }
    }

    /**
     * This method gets the worker associated with the local place via the
     * @link{de.uts.plm.ftglb.incremental.Worker#st st}-Attribute of
     * @link{de.uts.plm.ftglb.incremental.Worker Worker}-class.
     */
    private def st() : Worker[Queue, R]
    {
        return (this.worker.st());
    }
}