package de.uks.plm.ftglb.incremental.managers;

import x10.util.ArrayList;
import x10.util.List;
import x10.util.Collection;
import x10.util.Pair;

import de.uks.plm.ftglb.incremental.dataStructures.Timeout;

/**
 * This class manages timeouts and checks them, when a timeout becomes due
 * when the check-method is called.
 * In addition check() returns a List of dead places or null, if none are dead.
 */
public class TimeoutManager
{
    public val waitingFor:ArrayList[Timeout];

    public def this(val timeouts:Collection[Timeout]):TimeoutManager
    {
        this();
        this.waitingFor.addAll(timeouts);
    }

    public def this():TimeoutManager
    {
        this.waitingFor = new ArrayList[Timeout]();
    }

    /**
     * Add a timeout to the manager
     * Role: any
     * Sink: yes
     */
    public def add(val t:Timeout) : Boolean
    {
        if (Place(t.placeId).isDead())
        {
            return (false);
        }
        if (!this.waitingFor.contains(t))
        {
            this.waitingFor.add(t);
        }
        return (true);
    }

    /**
     * Check all timeouts.
     * Return a list of dead places.
     * Role: any
     * Sink: yes
     */
    public def check() : List[Timeout]
    {
        val deadPlaces:ArrayList[Timeout] = new ArrayList[Timeout]();
        this.waitingFor.sort();
        for(var index:Long = 0; index < this.waitingFor.size(); ++index)
        {
            val entry:Timeout = this.waitingFor(index);
            // List is sorted by the next Checktime of the Timeouts
            val now:Long = System.currentTimeMillis();
            if (now > entry.getNextCheck())
            {
                val isAlive = entry.check();
                if (!isAlive)
                {
                   deadPlaces.add(entry);
                   this.waitingFor.removeAt(index);
                   --index;
                }
            }
            else
            {
                break;
            }
        }

        if (0 < deadPlaces.size())
        {
            return (deadPlaces);
        }
        // else if (0 == deadPlaces.size())
        return (null);
    }

    /**
     * Remove a timeout from the manager.
     * Role: any
     * Sink: yes
     */
    public def remove(val place:Long, val reason:Long) : void
    {
        val timeout:Timeout = new Timeout(place, reason);
        this.waitingFor.remove(timeout);
    }

    public static def checkNow(val timeout:Timeout) : Boolean
    {
        timeout.setNextCheck(System.currentTimeMillis() + timeout.interval);
        val place:Place = Place(timeout.placeId);
        return (!place.isDead());
    }
}