package de.uks.plm.ftglb.incremental.utsgOneQueue;

import x10.compiler.*;
import de.uks.plm.utils.RingBuffer;

public class UTS extends RingBuffer[TreeNode]
{
    public var count:Long;
    public val den:Double;

    public def this(val factor:Int) : UTS
    {
        this(factor, 4096);
    }
    
    public def this(val factor:Int, val size:Long) : UTS
    {
        super(size);
        den = Math.log(factor / (1.0 + factor));
        this.count = 0;
    }

    public def this(val den:Double, val size:Long) : UTS
    {
        super(size);
        this.den = den;
        this.count = 0;
    }

    @Inline public final def init(val seed:Int, val height:Int) : void
    {
        push(SHA1Rand(seed, height));
        ++this.count;
    }

    @Inline public final def push(val h:SHA1Rand) : void
    {
        val u:Long
            = Math.floor(Math.log(1.0 - h() / 2147483648.0) / this.den) as Long;

        if (h.depth > 1n)
        {
            val newNodes:Rail[TreeNode] = new Rail[TreeNode](u);
            for (var i:Long = 0; i < u; ++i)
            {
                newNodes(i) = new TreeNode(h, i);
            }
            this.pushTop(newNodes);
        }
        else
        {
            this.count += u;
        }
    }

    @Inline public final def expand() : void
    {
        val h = this.popTop();
        val d = h.parent.depth;

        if(1n < d) { this.push(SHA1Rand(h.parent, h.id as Int, d - 1n)); }
    }

    static def sub(val str:String, val start:Int, val end:Int) : String
    {
        return (str.substring(start, Math.min(end, str.length())));
    }

    public static def print(val time:Long, val count:Long) : void
    {
        Console.OUT.println("Performance: " + count + "/"
                            + sub("" + time/1e9, 0n, 6n) + " = "
                            + sub("" + (count/(time/1e3)), 0n, 6n)
                            + "M nodes/s");
    }

    public static def main(val args:Rail[String]) : void
    {
        val queue = new UTS(4n); // branching factor
        Console.OUT.println("Starting...");
        var time:Long = System.nanoTime();
        queue.init(19n, 13n); // seed, depth
        while (queue.size() > 0)
        {
            queue.expand();
            ++queue.count;
        }
        time = System.nanoTime() - time;
        Console.OUT.println("Finished.");
        print(time, queue.count);
    }

    /**
     * Clone-constructor.
     */
    def this(val that:UTS) : UTS
    {
        super(that);
        this.den = that.den;
        this.count = that.count;
    }
}