package de.uks.plm.ftglb.incremental.utsgOneQueue;

import x10.compiler.*;

import x10.util.Option;
import x10.util.OptionsParser;
import x10.util.Random;

// import x10.xrx.Runtime;

import de.uks.plm.ftglb.incremental.*;

public final class UTSG
{
    public static def compute(val args:Rail[String]) : Rail[Long]
    {
       val opts:OptionsParser
           = new OptionsParser
               (args
                , new Rail[Option]()
                , [Option("b", "", "Branching factor")
                   , Option("r", "", "Seed (0 <= r < 2^31)")
                   , Option("d", "", "Tree depth")
                   , Option("n"
                            , ""
                            ,   "Number of nodes to process before probing."
                              + " Default 200.")
                   , Option("w"
                            , ""
                            , "Number of thieves to send out. Default 1.")
                   , Option("l", "", "Base of the lifeline")
                   , Option("m", "", "Max potential victims")
                   , Option("v", "", "Verbose. Default 0 (no).")
                   , Option("k"
                            , ""
                            ,  "Number of cycles, before a cyclic backup is "
                             + "written.")
                   , Option("g"
                            , ""
                            ,  "Number of cycles, before a monitor-call is "
                             + "started (all r processRecorde")]);

        val b:Int = opts("-b", 4n);
        val r:Int = opts("-r", 19n);
        val d:Int = opts("-d", 13n);
        val n:Int = opts("-n", 511n);
        val l:Int = opts("-l", 32n);
        val m:Int = opts("-m", 1024n);
        val k:Int = opts("-k", 2048n);
        val g:Int = opts("-g", 128n);
        val verbose:Int = opts("-v", GLBParameters.SHOW_RESULT_FLAG);

        val P:Long = Place.numPlaces();

        var z0:Int = 1n;
        var zz:Int = l;
        while (zz < P)
        {
            z0++;
            zz *= l;
        }
        val z:Int = z0;
        val w:Int = opts("-w", z);

        Console.OUT.println(  "places = " + P
                            + "   b = " + b
                            + "   r = " + r
                            + "   d = " + d
                            + "   w = " + w
                            + "   n = " + n
                            + "   l = " + l
                            + "   m = " + m
                            + "   z = " + z
                            + "   k = " + k
                            + "   g = " + g);

        val init:()=>Queue = ()=>{ return (new Queue(b)); };

        val glb:GLB[Queue, Long] = new GLB[Queue, Long](init
                                                        , GLBParameters
                                                            (n
                                                             , w
                                                             , l
                                                             , z
                                                             , m
                                                             , k
                                                             , g
                                                             , verbose)
                                                        , true );

        val start:()=>void = () =>
                             {
                                 glb.taskQueue().init(r, d);
                             };

        val result:Rail[Long] = glb.run(start);
        return (result);
    }

    public static def main(val args:Rail[String]) : void
    {
        compute(args);
    }

    public static def mainTest(val args:Rail[String]) : Rail[Long]
    {
        val result:Rail[Long] = compute(args);
        Console.OUT.println("Result is: " + result(0));
        return (result);
    }
}